#-----------------------------------------------------------------------------
# Version: 2.0.0
# Compatible: SketchUp 7.1 (PC)
#-----------------------------------------------------------------------------
#
# CHANGELOG
# See end of file
#
#-----------------------------------------------------------------------------
#
# TO DO
# * Transfer Group/Component Materials to Faces/Edges (?)
# * Adjust axis
# * Config dialog - live settings
# * Merge similar materials
# * Find similar groups/components
# * Add Plugin Manager support.
#
#-----------------------------------------------------------------------------
#
# Thomas Thomassen
# thomas[at]thomthom[dot]net
#
#-----------------------------------------------------------------------------

require 'sketchup.rb'

#-----------------------------------------------------------------------------


module TT_Cleanup
  
  
  # Improve. Make Settings class.
  def self.get_setting(setting, default)
    setting = setting.to_s if setting.is_a?(Symbol)
    Sketchup.read_default('TT_Cleanup', setting, default)
  end
  
  def self.set_setting(setting, value)
    setting = setting.to_s if setting.is_a?(Symbol)
    Sketchup.write_default('TT_Cleanup', setting, value)
  end
  
  def self.save_options
    @options.each { |k,v|
      self.set_setting(k, v)
    }
  end

  @options = {}
  @options[:merge_faces]       = self.get_setting(:merge_faces, true)
  @options[:ignore_normals]    = self.get_setting(:ignore_normals, false)
  @options[:ignore_materials]  = self.get_setting(:ignore_materials, false)
  @options[:ignore_uv]         = self.get_setting(:ignore_uv, false)
  @options[:remove_dups]       = self.get_setting(:remove_dups, true)
  @options[:repair_edges]      = self.get_setting(:repair_edges, true)
  @options[:lonely_edges]      = self.get_setting(:lonely_edges, true)
  @options[:remove_edge_mat]   = self.get_setting(:remove_edge_mat, false)
  @options[:smooth_angle]      = self.get_setting(:smooth_angle, nil)
  @options[:purge]             = self.get_setting(:purge, true)
  
  
  unless file_loaded?('tt_cleanup.rb')
    parent_menu = ($tt_menu) ? $tt_menu : UI.menu('Plugins')
    menu = parent_menu.add_submenu('Cleanup')
    menu.add_item('Cleanup Entire Model')  { self.cleanup! }
    menu.add_item('Cleanup Local Context')  { self.cleanup_local! }
    menu.add_separator
    
    item = menu.add_item('Merge Faces') {
      @options[:merge_faces] = !@options[:merge_faces]
      self.save_options
    }
    menu.set_validation_proc(item)      {
      (@options[:merge_faces]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    menu_strict = menu.add_item(' Ignore Normals') {
      @options[:ignore_normals] = !@options[:ignore_normals]
      self.save_options
    }
    menu.set_validation_proc(menu_strict)      {
      (@options[:ignore_normals]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    menu_ignore_mat = menu.add_item(' Ignore Materials') {
      @options[:ignore_materials] = !@options[:ignore_materials]
      self.save_options
      
    }
    menu.set_validation_proc(menu_ignore_mat)  {
      (@options[:ignore_materials]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    menu_ignore_uv  = menu.add_item(' Ignore UV') {
      @options[:ignore_uv] = !@options[:ignore_uv]  
      self.save_options
    }
    menu.set_validation_proc(menu_ignore_uv)   {
      (@options[:ignore_uv]) ? MF_CHECKED : MF_UNCHECKED
    }

    menu.add_separator
    
    menu_remove_dups = menu.add_item('Erase Duplicate Faces') {
      @options[:remove_dups] = !@options[:remove_dups]
      self.save_options
    }
    menu.set_validation_proc(menu_remove_dups)  {
      (@options[:remove_dups]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    menu_repair_edges = menu.add_item('Repair Split Edges') {
      @options[:repair_edges] = !@options[:repair_edges]
      self.save_options
    }
    menu.set_validation_proc(menu_repair_edges)  {
      (@options[:repair_edges]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    menu_lone_edges = menu.add_item('Erase Lonely Edges') {
      @options[:lonely_edges] = !@options[:lonely_edges]
      self.save_options
    }
    menu.set_validation_proc(menu_lone_edges)  {
      (@options[:lonely_edges]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    remove_edge_mat = menu.add_item('Remove Edge Material') {
      @options[:remove_edge_mat] = !@options[:remove_edge_mat]
      self.save_options
    }
    menu.set_validation_proc(remove_edge_mat)  {
      (@options[:remove_edge_mat]) ? MF_CHECKED : MF_UNCHECKED
    }
    
    remove_edge_mat = menu.add_item('Set Edge Smooth') {
      angle = (@options[:smooth_angle].nil?) ? '--' : @options[:smooth_angle].to_s
      input = UI.inputbox(['Smooth Edges by Angle : '], [angle], 'Edge Smooth Angle')
      if input
        @options[:smooth_angle] = (input[0].match(/^\d+(\.\d+)?$/).nil?) ? nil : input[0].to_f
        self.save_options
      end
    }
    
  end


  # Process the whole model
  def self.cleanup!
    # Any version less than 7.1 will be warned about unpredictable results.
    msg = 'Sketchup prior to 7.1 has a bug which might lead to loss of geometry. Do you want to continue?'
    return if self.version_check(7, 1, msg) == 7 # No
    
    prompts = [
      'Merge Coplanar Faces: ',
      '> Ignore Normals: ',
      '> Ignore Materials: ',
      '> Ignore UV: ',
      'Remove Duplicate Faces: ',
      'Repair Split Edges: ',
      'Remove Lonely Edges: ',
      'Remove Edge Materials: ',
      'Smooth Edges by Angle : ',
      'Purge: '
      ]
    
    defaults = [
      (@options[:merge_faces]) ? 'Yes' : 'No',
      (@options[:ignore_normals]) ? 'Yes' : 'No',
      (@options[:ignore_materials]) ? 'Yes' : 'No',
      (@options[:ignore_uv]) ? 'Yes' : 'No',
      (@options[:remove_dups]) ? 'Yes' : 'No',
      (@options[:repair_edges]) ? 'Yes' : 'No',
      (@options[:lonely_edges]) ? 'Yes' : 'No',
      (@options[:remove_edge_mat]) ? 'Yes' : 'No',
      (@options[:smooth_angle].nil?) ? '--' : @options[:smooth_angle].to_s,
      (@options[:purge]) ? 'Yes' : 'No'
    ]
    
    lists = [
      'Yes|No',
      'Yes|No',
      'Yes|No',
      'Yes|No',
      'Yes|No',
      'Yes|No',
      'Yes|No',
      'Yes|No',
      '',
      'Yes|No'
    ]
    input = UI.inputbox(prompts, defaults, lists, 'Clean Up Model')
    return if input == false
    
    #puts input
    
    @options[:merge_faces]      = (input[0] == 'Yes') ? true : false
    @options[:ignore_normals]   = (input[1] == 'Yes') ? true : false
    @options[:ignore_materials] = (input[2] == 'Yes') ? true : false
    @options[:ignore_uv]        = (input[3] == 'Yes') ? true : false
    @options[:remove_dups]      = (input[4] == 'Yes') ? true : false
    @options[:repair_edges]     = (input[5] == 'Yes') ? true : false
    @options[:lonely_edges]     = (input[6] == 'Yes') ? true : false
    @options[:remove_edge_mat]  = (input[7] == 'Yes') ? true : false
    @options[:smooth_angle]     = (input[8].match(/^\d+(\.\d+)?$/).nil?) ? nil : input[8].to_f
    @options[:purge]            = (input[9] == 'Yes') ? true : false
    self.save_options
    
    model = Sketchup.active_model
    
    model.start_operation('Cleanup Model', true)
    
    stats = {}
    stats['Total Elapsed Time'] = Time.now	
    
    # --- Purging --- #
    if @options[:purge]
      stats['Purged Components'] = model.definitions.length
      Sketchup.status_text = 'Purging Components...'
      model.definitions.purge_unused
      stats['Purged Components'] -= model.definitions.length
    end
    
    # --- Fix Duplicate Component Names --- #
    comps = self.fix_comp_names
    if comps > 0
      puts "> Fixed Duplicate Component Names: #{comps}"
      stats['Duplicate Component Names Fixed'] = comps
    end
    
    # --- Merge Coplanar Faces --- #
    if @options[:merge_faces] 
      stats['Edges Reduced'] = 0
      stats['Faces Reduced'] = model.number_faces if model.respond_to?(:number_faces)
      edges = self.merge_faces(model.entities, @options)
      # Model context
      stats['Edges Reduced'] += edges
      # Definition context
      model.definitions.each { |d|
        next if d.image?
        edges = self.merge_faces(d.entities, @options)
        stats['Edges Reduced'] += edges
      }
      stats['Faces Reduced'] -= model.number_faces if model.respond_to?(:number_faces)
    end
    
    # --- Erase Duplicate Faces --- #
    if @options[:remove_dups]
      stats['Edges Reduced'] ||= 0
      stats['Faces Reduced'] ||= 0
      # Model context
      faces = self.erase_duplicate_faces(model.entities)
      stats['Faces Reduced'] += faces
      # Definition context
      model.definitions.each { |d|
        next if d.image?
        faces = self.erase_duplicate_faces(d.entities)
        stats['Edges Reduced'] += edges
      }
      
      # Merge Coplanar Faces
      if @options[:merge_faces] 
        faces = 0
        faces = model.number_faces if model.respond_to?(:number_faces)
        edges = self.merge_faces(model.entities, @options)
        # Model context
        stats['Edges Reduced'] += edges
        # Definition context
        model.definitions.each { |d|
          next if d.image?
          edges = self.merge_faces(d.entities, @options)
          stats['Edges Reduced'] += edges
        }
        faces -= model.number_faces if model.respond_to?(:number_faces)
        stats['Faces Reduced'] += faces
      end
      
    end
    
    # --- Repair Split Edges --- #
    if @options[:lonely_edges] 
      stats['Edges Reduced'] ||= 0
      edges = self.erase_lonely_edges(model.entities)
      stats['Edges Reduced'] += edges
      model.definitions.each { |d|
        next if d.image?
        edges = self.erase_lonely_edges(d.entities)
        stats['Edges Reduced'] += edges
      }
    end
    
    # --- Repair Split Edges --- #
    if @options[:repair_edges] 
      edges = self.repair_edges(model.entities)
      stats['Edges Repaired'] = edges
      model.definitions.each { |d|
        next if d.image?
        edges = self.repair_edges(d.entities)
        stats['Edges Repaired'] += edges
      }
    end
    
    # --- Process the rest --- #
    "Processing Model. Please wait..."
    model.entities.each { |e|
      self.process_entity(e, @options)
    }
    model.definitions.each { |d|
      Sketchup.status_text = "Processing #{d.name}. Please wait..."
      d.entities.each { |e|
        self.process_entity(e, @options)
      }
    }
    
    # --- Purging --- #
    if @options[:purge]
      stats['Purged Layers'] = model.layers.length
      Sketchup.status_text = 'Purging Layers...'
      model.layers.purge_unused
      stats['Purged Layers'] -= model.layers.length
      
      stats['Purged Materials'] = model.materials.length
      Sketchup.status_text = 'Purging Materials...'
      model.materials.purge_unused
      stats['Purged Materials'] -= model.materials.length
      
      stats['Purged Styles'] = model.styles.count
      Sketchup.status_text = 'Purging Styles...'
      model.styles.purge_unused
      stats['Purged Styles'] -= model.styles.count
    end
    
    stats['Total Elapsed Time'] = Time.now - stats['Total Elapsed Time']
    
    
    Sketchup.status_text = 'Done!'
    model.commit_operation 
    
    # Stats
    UI.messagebox("Cleanup Statistics:\n#{stats.collect{|k,v|"> #{k}: #{v}"}.sort.join("\n")}", MB_MULTILINE)
    self.validity_check
  end
  
  
  def self.process_entity(e, options)
    if e.is_a?(Sketchup::Edge)
      # Remove Edge Material
      e.material = nil if options[:remove_edge_mat]
      # Smooth Edge
      if !options[:smooth_angle].nil? && e.faces.length == 2
        angle = e.faces[0].normal.angle_between(e.faces[1].normal)
        if angle.radians.abs <= options[:smooth_angle]
          e.smooth = true
          e.soft = true
        end
      end
    end
  end
  
  
  # Find all coplanar edges in the selection and erase them.
  # The order of the clean up process is very important in order to fix all
  # problems.
  def self.cleanup_local!
    # Any version less than 7.1 will be warned about unpredictable results.
    msg = 'Sketchup prior to 7.1 has a bug which might lead to loss of geometry. Do you want to continue?'
    return if self.version_check(7, 1, msg) == 7 # No
    
    model = Sketchup.active_model
    sel = model.selection
    model.start_operation('Cleanup', true)
    
    # If nothing is selected - pick the current context
    ents = (sel.empty?) ? model.active_entities : sel.to_a
    
    puts "Cleanup"
    
    # --- Fix Duplicate Component Names --- #
    comps = self.fix_comp_names
    puts "> Fixed Duplicate Component Names: #{comps}" if comps > 0
    
    # --- Merge Coplanar Faces --- #
    if @options[:merge_faces]
      edges = self.merge_faces(ents, @options)
      puts "> Edges Erased: #{edges}"
    end
    
    # --- Erase Duplicate Faces --- #
    # Faces are merged first to reduce the processing time. (Less faces)
    # After removing duplicates the remaining faces needs to be merged.
    if @options[:remove_dups] 
      faces = self.erase_duplicate_faces(ents)
      puts "> Duplicate Faces Erased: #{faces}"
      
      if @options[:merge_faces]
        edges = self.merge_faces(ents, @options)
        puts "> Edges Erased: #{edges}"
      end
    end
    
    # --- Erase Lonely Edge --- #
    if @options[:lonely_edges]
      edges = self.erase_lonely_edges(ents)
      puts "> Lonely Edges Erased: #{edges}"
    end
    
    # --- Repair Split Edges --- #
    if @options[:repair_edges] 
      edges = self.repair_edges(ents)
      puts "> Split Edges Repaired: #{edges}"
    end
    
    # --- Process the rest --- #
    Sketchup.status_text = "Processing. Please wait..."
    for e in ents
      self.process_entity(e, @options)
    end
    
    Sketchup.status_text = "Cleanup complete!"
    model.commit_operation
    
    self.validity_check
  end
  
  
  # merge faces
  def self.merge_faces(entities, options)   
    Sketchup.status_text = 'Merging faces. Please wait...'
    
    c = 0
    #p options
    # Merge faces first. There might be additional edges after such merge, so lonely edges
    # are checked at the end.
    entities.to_a.each { |e|
      next unless e.valid? && e.is_a?(Sketchup::Edge)
      # Coplanar edges only have two faces connected.
      next unless e.faces.length == 2
      f1 = e.faces[0]
      f2 = e.faces[1]
      # Don't try to merge faces sharing the same set of vertices.
      next if face_duplicate?(f1, f2)
      # We might want to consider the materials applied before removing any edge.
      unless options[:ignore_materials]
        if f1.material == f2.material && f1.back_material == f2.back_material
          unless f1.material.nil? || f1.material.texture.nil? || options[:ignore_uv]
            tw = Sketchup.create_texture_writer
            uvh1 = f1.get_UVHelper(true, true, tw)
            uvh2 = f2.get_UVHelper(true, true, tw)
            p1 = e.start.position
            p2 = e.end.position
            # Can/Need this improved (?)
            next unless uvh1.get_front_UVQ(p1) == uvh2.get_front_UVQ(p1) &&
                        uvh1.get_front_UVQ(p2) == uvh2.get_front_UVQ(p2) &&
                        uvh1.get_back_UVQ(p1)  == uvh2.get_back_UVQ(p1) &&
                        uvh1.get_back_UVQ(p2)  == uvh2.get_back_UVQ(p2)
          end # unless options[:ignore_uv]
        else
          next
        end
      end # unless options[:ignore_materials]
      # 
      if options[:ignore_normals] 
        if self.faces_coplanar?(f1, f2)
          e.erase!
          c += 1
        end
      else
        if self.faces_coplanar?(f1, f2) && f1.normal.samedirection?(f2.normal)
          e.erase!
          c += 1
        end
      end
    } # each
    c
  end
  
  
  # Erase edges not connected to faces,
  # and edges that connects to the same face multiple times.
  def self.erase_lonely_edges(entities)
    Sketchup.status_text = 'Hunting for redundant edges. Please wait...'
    
    return 0 if entities.length == 0
    entities = entities.select { |e| e.valid? }
    parent = entities[0].parent.entities
    
    edges = []
    entities.to_a.each { |e|
      next unless e.valid? && e.is_a?(Sketchup::Edge)
      if e.faces.length == 0 || 
         ( e.faces.length > 1 && e.faces.all?{ |f| f == e.faces[0] } )
        edges << e
      end
    }
    parent.erase_entities(edges)
    return edges.length
  end
  
  
  # Finds multiple faces for the same set of vertices and reduce them to one.
  # Erases faces overlapped by a larger face.
  def self.erase_duplicate_faces(entities)
    Sketchup.status_text = "Removing duplicate faces..."
    
    return 0 if entities.length == 0
    entities = entities.select { |e| e.valid? }
    parent = entities[0].parent.entities
    
    faces = entities.select { |e| e.is_a?(Sketchup::Face) }
    duplicates = []
    
    #p faces
    #p entities
    
    for face in faces
      next unless face.valid?
      #puts ' '
      #p face
      #puts face.area
      next if duplicates.include?(face)
      connected = face.edges.map { |e| e.faces }
      connected.flatten!
      connected.uniq!
      connected &= entities
      connected.delete(face)
      dups = false
      #puts "> Connected: #{connected.inspect}"
      for f in (connected - duplicates)
        next unless f.valid?
        #puts "> #{f.inspect} - #{face_duplicate?(face, f, true)}"
        duplicates << f if face_duplicate?(face, f, true)
        #f.erase! if face_duplicate?(face, f, true)
      end # for
    end
    parent.erase_entities(duplicates) unless duplicates.empty?
    #Sketchup.active_model.selection.clear
    #Sketchup.active_model.selection.add(duplicates)
    
    return duplicates.length
  end
  
  
  # Based on repair_broken_lines.rb by Carlo Roosen 2004
  def self.repair_edges(entities)
    Sketchup.status_text = "Repairing split edges..."
    
    # This method makes temporary edges which should not be split - so the Break Edges feature
    # in SU7+ is disabled for the duration of this method.
    re = Sketchup.break_edges?
    Sketchup.break_edges = false
    
    edges = []
    for e in entities
      next unless e.is_a?(Sketchup::Edge)
      for v in e.vertices
        next unless v.edges.length == 2
        v1 = v.edges[0].line[1]
        v2 = v.edges[1].line[1]
        next unless v1.parallel?(v2)
        # To repair a broken edge a temporary edge is placed at their shared vertex.
        # This temporary edge is then later erased which causes the two edges to
        # merge.
        edge = entities.add_line( v.position, v.position.offset(v1.axes.x) )
        edges << edge unless edge.nil?
      end
    end
    entities.erase_entities(edges) unless edges.empty?
    
    ensure
    # Restore the break_edges setting
    Sketchup.break_edges = re
    return edges.length
  end
  
  
  def self.face_duplicate?(face1, face2, overlapping = false)
    return false if face1 == face2
    v1 = face1.outer_loop.vertices
    v2 = face2.outer_loop.vertices
    return true if (v1 - v2).empty? && (v2 - v1).empty?
    #return true if overlapping && (v2 - v1).empty? # (!) error
    # A wee hack to determine if a face2 is fully overlapped by face1.
    if overlapping && (v2 - v1).empty?
      edges = (face2.outer_loop.edges - face1.outer_loop.edges)
      unless edges.empty?
        point = edges[0].start.position.offset(edges[0].line[1], 0.01)
        return true if face1.classify_point(point) <= 4
      end
    end
    return false
  end
  
  
  def self.faces_coplanar?(face1, face2)
    vs = face1.vertices + face2.vertices
    plane = Geom.fit_plane_to_points(vs)
    vs.each { |v| 
      return false unless v.position.on_plane?(plane)
    }
    return true
  end
  
  
  def self.fix_comp_names
    Sketchup.status_text = "Looking for multiple components of the same name..."
    
    model = Sketchup.active_model
    c = 0
    for definition in model.definitions
      copies = model.definitions.select { |d|
        d != definition && d.name == definition.name
      }
      next if copies.empty?
      puts "> Multiple definitions for '#{definition.name}' found!"
      for copy in copies
        puts "  > Renaming '#{copy.name}' to '#{model.definitions.unique_name(copy.name)}'..."
        copy.name = model.definitions.unique_name(copy.name)
        c += 1
      end
    end
    c
  end
  
  
  def self.validity_check
    Sketchup.send_action(21124)
  end
  
  
  # Version Check
  def self.version_check(major, minor, message)
    v = Sketchup.version.split('.').map! { |i| i.to_i }
    return if v[0] >= major
    return if v[0] == major && v[1] >= minor
    return UI.messagebox(message, MB_YESNO)
  end

end


#-----------------------------------------------------------------------------
file_loaded('tt_cleanup.rb')
#-----------------------------------------------------------------------------
#
# CHANGELOG
# 1.0.0 - 21.10.2009
#    * Simple search for coplanar edges.
#
# 1.0.1 - 22.10.2009
#    * Decreased the tolerance.
#    * Added option to respect materials.
#
# 1.1.0 - 23.10.2009
#    * Cleanup entire model.
#      > Erase all coplanar edges
#      > Purge
#      > Remove Edge Materials
#      > Smooth Edges
#
# 1.1.1 - 23.10.2009
#    * Added option to erase lonely edges.
#
# 1.1.2 - 23.10.2009
#    * Added option to erase lonely edges.
#
# 1.2.0 - 23.04.2010
#    * Improved method to detect co-planar faces.
#
# 1.2.1 - 23.04.2010
#    * Minor bug-fixes.
#
# 1.2.2 - 24.04.2010
#    * Reworked how the plugin erases edges between coplanar faces.
#
# 1.2.3 - 25.04.2010
#    * Fixed bug regarding reference to deleted entities.
#
# 1.3.0 - 26.04.2010
#    * Added option to make different UV mapping prevent faces from merging.
#
# 2.0.0 - 05.05.2010
#    * Code cleanup.
#    * Persistent preferences.
#    * Settings for global and local cleanup synced.
#    * Repairs split edges.
#    * Removes duplicate faces.
#    * Fixes duplicate component definition names.
#    * Runs Validity Check at the end.
#
#-----------------------------------------------------------------------------